#!/usr/bin/env python
# coding: utf-8

# # import libraries
import pandas as pd
import numpy as np
import os
import time
import platform
from sklearn.preprocessing import (
    StandardScaler,
    MinMaxScaler,
    MaxAbsScaler,
    RobustScaler,
    QuantileTransformer,
    PowerTransformer,
    Normalizer,
)
from sklearn.preprocessing import LabelEncoder, OneHotEncoder
from sklearn.metrics import (
    accuracy_score,
    precision_score,
    recall_score,
    f1_score,
    cohen_kappa_score,
    roc_auc_score,
)
from tensorflow.keras.layers import (
    Dense,
    Conv2D,
    MaxPooling2D,
    Flatten,
    Dropout,
    BatchNormalization,
    Input,
    Add,
    Activation,
)
from tensorflow.keras.models import Model, load_model, save_model
from sklearn.model_selection import train_test_split, ShuffleSplit
from tensorflow.keras.callbacks import ModelCheckpoint, LearningRateScheduler
from matplotlib import pyplot

# ## functions
# ### dataset loader function
def dataset_loader(path="", full=True, specific="", verbose=1):
    if path == "":
        print("\nset path to default:", "\ndataset build/UAHdataset.csv\n")
        data = pd.read_csv("dataset build/UAHdataset.csv")
    else:
        try:
            data = pd.read_csv(path)
        except Exception:
            print("\nERROR: bad path entry", "\nReturn empty data")
            print("\nNote: pandas must be imported as pd")
            data = []
            return data
    while full == True:
        data_name = "full data loaded"
        return data
    specific = str(specific)
    if specific == "secondary road" or specific == "":
        data = data.loc[data["road"] == "secondary"]
        data = data.drop("road", axis=1)
        data_name = "secondary road data loaded"
    elif specific == "motorway road" or specific == "0":
        data = data.loc[data["road"] == "motorway"]
        data = data.drop("road", axis=1)
        data_name = "motorway road data loaded"
    elif int(specific) < 7:
        data = data.loc[data["driver"] == int(specific)]
        data = data.drop("driver", axis=1)
        data_name = "driver specific data loaded" + "\ndriver number:" + specific + "\n"
    else:
        print(
            "ERROR: wrong specific entry or specific entry does not exist",
            "\nEmpty data returned ",
        )
        print(
            'Note: Pandas library needed for this function to work and must be loaded as "pd"'
        )
        data = []
    if verbose == 1:
        print(data_name)
    return data


# ### label encoder function
def encoding_label(data, target, verbose=1):
    encoder = LabelEncoder()
    df = pd.DataFrame(data)
    try:
        if verbose == 1:
            print("label encoder target: ", data[target].unique())
        df[target] = encoder.fit_transform(data[target])
        if verbose == 1:
            print("target after label encoding : ", df[target].unique())
    except Exception:
        if verbose == 1:
            print(
                f"ERROR: target name '{target}' is not available in data\n",
                f"no label encoding realized for '{target}'\n",
            )
        return data
    return df


# ### data normalization function
def sklearn_data_normalizer(data, excluded_axis=[], scaler="minmax", verbose=1):
    normalized = data
    for i in excluded_axis:
        normalized = normalized.drop(i, axis=1)
    reserved_columns = normalized.columns
    scaler = str(scaler)
    if scaler == "0" or scaler == "raw_data":
        normalizer_name = "RawData (no scaling)"
        normalized = pd.DataFrame()
        for column in data.columns:
            normalized[column] = data[column].astype("float")
            normalized = normalized.reset_index(drop=True)
        normalized = normalized.fillna(0)
        if verbose == 1:
            print("data is not normalized, returned:", normalizer_name)
        return normalized, normalizer_name
    elif scaler == "1" or scaler == "minmax":
        scalerfunction = MinMaxScaler()
        normalizer_name = "MinMaxscaler"
    elif scaler == "2" or scaler == "standard":
        scalerfunction = StandardScaler()
        normalizer_name = "Standardscaler"
    elif scaler == "3" or scaler == "maxabs":
        scalerfunction = MaxAbsScaler()
        normalizer_name = "MaxAbsScaler"
    elif scaler == "4" or scaler == "robust":
        scalerfunction = RobustScaler()
        normalizer_name = "RobustScaler"
    elif scaler == "5" or scaler == "quantile_normal":
        scalerfunction = QuantileTransformer(output_distribution="normal")
        normalizer_name = "QuantileTransformer using normal distribution"
    elif scaler == "6" or scaler == "quantile_uniform":
        scalerfunction = QuantileTransformer(output_distribution="uniform")
        normalizer_name = "QuantileTransformer using uniform distribution"
    elif scaler == "7" or scaler == "power_transform":
        scalerfunction = PowerTransformer(method="yeo-johnson")
        normalizer_name = "PowerTransformer using method yeo-johnson"
    elif scaler == "8" or scaler == "normalizer":
        scalerfunction = Normalizer()
        normalizer_name = " sklearn Normalizer"
    else:
        print("\nERROR: wrong entry or wrong scaler type")
        print("\nreturned input data")
        return data
    if verbose == 1:
        print("data normalized with the", normalizer_name)
    normalized = scalerfunction.fit_transform(normalized)
    normalized = pd.DataFrame(normalized, columns=reserved_columns)
    for i in excluded_axis:
        normalized[i] = data[i]
    normalized = normalized.fillna(0)
    return normalized, normalizer_name


# ### One hot encoder function
def One_hot_encoding(data, target="target", verbose=1):
    encoder = OneHotEncoder()
    try:
        if verbose == 1:
            print("One Hot Encoder target: ", data[target].unique())
        encoded = encoder.fit_transform(data[target].values.reshape(-1, 1)).toarray()
    except Exception:
        try:
            if verbose == 1:
                print("One Hot Encoder target: ", data.unique())
            encoded = encoder.fit_transform(data.values.reshape(-1, 1)).toarray()
        except Exception:
            if verbose == 1:
                print(
                    f"ERROR: target name '{target}' is not available in data\n",
                    f"no One hot encoding realized for '{target}'\n",
                )
            return data
    if verbose == 1:
        print("example of target after One Hot encoding : ", encoded[0])
    df = pd.DataFrame(encoded)
    return df


# ### create Deep RCN model function
def create_resnet_model(input_shape=(20, 1, 1), class_nb=3):
    inputs = Input(shape=input_shape, name="input")
    x = Conv2D(60, (4, 4), activation="relu", padding="same", name="conv1_60_4x4")(
        inputs
    )
    x = BatchNormalization(name="batch_norm1")(x)
    x = MaxPooling2D((2, 2), padding="same", name="Max_pool1_2x2")(x)
    x = Dropout(0.2, name="dropout1_0.2")(x)
    x = resnet_block(x, 60, 4)
    x = Conv2D(30, (4, 4), activation="relu", padding="same", name="conv2_30_4x4")(x)
    x = BatchNormalization(name="batch_norm2")(x)
    x = MaxPooling2D((2, 2), padding="same", name="Max_pool2_2x2")(x)
    x = Dropout(0.2, name="dropout2_0.2")(x)
    x = Conv2D(15, (4, 4), activation="relu", padding="same", name="conv3_15_4x4")(x)
    x = BatchNormalization(name="batch_norm3")(x)
    x = MaxPooling2D((2, 2), padding="same", name="Max_pool3_2x2")(x)
    x = Dropout(0.2, name="dropout3_0.2")(x)
    x = Conv2D(8, (4, 4), activation="relu", padding="same", name="conv4_8_4x4")(x)
    x = BatchNormalization(name="batch_norm4")(x)
    x = MaxPooling2D((1, 1), padding="same", name="Max_pool4_1x1")(x)
    x = Flatten(name="Flatten_layer")(x)
    x = Dense(units=224, input_dim=448, activation="relu", name="dense1_448x224")(x)
    x = BatchNormalization(name="batch_norm5")(x)
    outputs = Dense(units=class_nb, activation="softmax", name="dense2_224x10")(x)
    your_model = Model(inputs, outputs, name="your_Conv_model")
    return your_model


def resnet_block(input_data, filters, conv_size):
    x = Conv2D(
        filters, conv_size, activation="relu", padding="same", name="resnet_block1"
    )(input_data)
    x = BatchNormalization(name="resnet_block2")(x)
    x = Conv2D(
        filters, conv_size, activation=None, padding="same", name="resnet_block3"
    )(x)
    x = BatchNormalization(name="resnet_block4")(x)
    x = Dropout(0.2)(x)
    x = Add(name="resnet")([x, input_data])
    x = Activation("relu", name="resnet_activation")(x)
    return x


# ### preprocess the dataset function
def data_preprocessing(
    data, target_name="target", normalization="1", raw_data_entry=True, verbose=1
):
    if raw_data_entry == True:
        data_numeric = encoding_label(
            encoding_label(data, "target", verbose=verbose), "road", verbose=verbose
        ).drop("Timestamp (seconds)", axis=1)
        df, normalizer_name = sklearn_data_normalizer(
            data_numeric, [target_name], normalization, verbose=verbose
        )
    elif raw_data_entry == False:
        df = data
    X = df.drop(target_name, axis=1)
    Y = df[target_name]
    X = np.reshape(X.values, (len(X), X.shape[1], 1, 1))
    Y_origin = df["target"]
    target = df["target"]
    Y = One_hot_encoding(target)
    if verbose == 1:
        print(
            "X shape is :",
            X.shape,
            "\nY shape is :",
            Y.shape,
            "\nY_orgin shape is:",
            Y_origin.shape,
        )
    return X, Y, Y_origin


# ### learning rate custom stepping function
def step_adam(epoch, scheduler_threshold=480, initial_lrate=0.001, second_lrate=0.0001):
    lrate = 0.001
    if epoch > 480:
        lrate = 0.0001
        print("Change in the learning rate, the new learning rate is:", lrate)
    return lrate


# ### model training function
def train_model(
    X,
    Y,
    Y_origin,
    model,
    epochs=600,
    batch_size=60,
    callbacks="best model",
    test_size=0.3,
    opt="adam",
    loss="categorical_crossentropy",
    metrics=["accuracy"],
    learning_rate="scheduler",
    scheduler_threshold=480,
    verbose=1,
):
    if callbacks == "best model":
        mc = ModelCheckpoint(
            "best_model.h5", monitor="val_accuracy", save_best_only=True, verbose=2
        )
        if any(c.isdigit() for c in platform.node()) == True:
            mc = ModelCheckpoint(
                "/content/drive/MyDrive/best_model.h5",
                monitor="val_accuracy",
                save_best_only=True,
                verbose=2,
            )
        cb = [mc]
        if learning_rate == "scheduler":
            lrs = LearningRateScheduler(step_adam)
            cb = [mc, lrs]
    elif callbacks == None:
        cb = None
    model.compile(loss=loss, optimizer=opt, metrics=metrics)
    X_train, X_test, Y_train, Y_test = train_test_split(
        X, Y, test_size=test_size, random_state=10
    )
    _, _, Yo_train, Yo_test = train_test_split(
        X, Y_origin, test_size=test_size, random_state=10
    )
    start_time = time.time()
    history = model.fit(
        X_train,
        Y_train,
        validation_data=(X_test, Y_test),
        callbacks=cb,
        epochs=epochs,
        batch_size=batch_size,
        verbose=verbose,
    )
    end_time = time.time()
    training_history = history.history
    save_history = pd.DataFrame(training_history).to_csv(
        r"training_history.csv", index=False
    )
    if any(c.isdigit() for c in platform.node()) == True:
        save_history = pd.DataFrame(training_history).to_csv(
            r"/content/drive/MyDrive/training_history.csv", index=False
        )
    if callbacks == "best model":
        try:
            best_model = load_model("/content/drive/MyDrive/best_model.h5")
        except Exception:
            try:
                best_model = load_model("best_model.h5")
            except Exception:
                print(
                    "exception triggered, can't load the saved best model, the best model is the current one"
                )
                best_model = model
    pred = model.predict(X_test, batch_size=batch_size, verbose=1)
    predictions = np.argmax(pred, axis=1)
    train_eval = model.evaluate(X_train, Y_train, verbose=verbose)
    train_score = train_eval[1] * 100
    exec_time = str(format((end_time - start_time), ".2f")) + " (s)"
    acc_score = accuracy_score(Yo_test.values, predictions, normalize=True) * 100
    pres_score = precision_score(Yo_test.values, predictions, average="macro") * 100
    rec_score = recall_score(Yo_test.values, predictions, average="macro") * 100
    f1 = f1_score(Yo_test.values, predictions, average="macro") * 100
    cokap_score = cohen_kappa_score(Yo_test.values, predictions) * 100
    roc_auc = roc_auc_score(Y_test.values, pred) * 100
    scores = [
        "execution time: {}".format(exec_time),
        "train accuracy: %.4f%%" % train_score,
        "accuracy: %.4f%%" % acc_score,
        "precision: %.4f%%" % pres_score,
        "recall: %.4f%%" % rec_score,
        "F1 score: %.4f%%" % f1,
        "cohen kappa: %.4f%%" % cokap_score,
        "roc_auc_score: %.4f%%" % roc_auc,
    ]
    if verbose == 1:
        print(scores)
    return model, best_model, training_history, scores


# ### cros validation evaluation function
def cros_validation_evaluation(
    dataset_path,
    model_fn="create_resnet_model",
    normalization=0,
    target_name="target",
    nb_splits=5,
    epochs=600,
    batch_size=60,
    callbacks="best model",
    test_size=0.3,
    opt="adam",
    loss="categorical_crossentropy",
    metrics=["accuracy"],
    learning_rate="scheduler",
    scheduler_threshold=120,
    dataset=None,
    verbose=1,
):
    if type(model_fn) == str:
        model_fn = model_fn + "()"
    else:
        model_fn = str(model_fn) + "()"
    if dataset_path == "" or dataset_path == None:
        if dataset == None:
            print("Empty dataset path and empty dataset entry")
            return None
        data = dataset
    else:
        data = dataset_loader(dataset_path)
    data_numeric = encoding_label(
        encoding_label(data, "target", verbose=verbose), "road", verbose=verbose
    ).drop("Timestamp (seconds)", axis=1)
    df, normalizer_name = sklearn_data_normalizer(
        data_numeric, [target_name], normalization, verbose=verbose
    )
    X_origin = df.drop(target_name, axis=1)
    Y_origin = df[target_name]
    cv = ShuffleSplit(n_splits=nb_splits, test_size=test_size, random_state=10)
    (
        exec_time,
        train_acc_scores,
        acc_scores,
        pres_scores,
        rec_scores,
        f1,
        cokap_scores,
        roc_auc_scores,
    ) = ([], [], [], [], [], [], [], [])
    cv_scores = pd.DataFrame()
    for train, test in cv.split(X_origin, Y_origin):
        print(
            "\033[1m"
            + "\n"
            + "*******"
            + "begin cross validation in fold number:{}".format(
                len(train_acc_scores) + 1
            )
            + "*******"
            + "\033[0m"
        )
        X_train = np.reshape(
            X_origin.loc[train].values,
            (len(X_origin.loc[train]), X_origin.loc[train].shape[1], 1, 1),
        )
        X_test = np.reshape(
            X_origin.loc[test].values,
            (len(X_origin.loc[test]), X_origin.loc[test].shape[1], 1, 1),
        )
        Y_train = One_hot_encoding(Y_origin[train], verbose=0)
        Y_test = One_hot_encoding(Y_origin[test], verbose=verbose)
        if callbacks == "best model":
            mc = ModelCheckpoint(
                "cross_validation/cv_best_model.h5",
                monitor="val_accuracy",
                save_best_only=True,
                verbose=2,
            )
            cb = [mc]
            if learning_rate == "scheduler":
                lrs = LearningRateScheduler(step_adam)
                cb = [mc, lrs]
        elif callbacks == None:
            cb = None
        try:
            cv_model = eval(model_fn)
        except Exception:
            print(
                "Error: model function variable is bad entry, please pass the name of your model building function as string in the variable: model_fn"
            )
            return
        newpath = "cross_validation"
        if not os.path.exists(newpath):
            os.makedirs(newpath)
        cv_model.compile(loss=loss, optimizer=opt, metrics=metrics)
        start_time = time.time()
        history = cv_model.fit(
            X_train,
            Y_train,
            validation_data=(X_test, Y_test),
            callbacks=cb,
            epochs=epochs,
            batch_size=batch_size,
            verbose=verbose,
        )
        end_time = time.time()
        training_history = history.history
        save_history = pd.DataFrame(training_history).to_csv(
            r"cross_validation/training_history{}.csv".format(
                len(train_acc_scores) + 1
            ),
            index=False,
        )
        if callbacks == "best model":
            try:
                best_model = load_model("cross_validation/cv_best_model.h5")
            except Exception:
                print(
                    "exception triggered, can't load the saved best model, the best model is the current one"
                )
                best_model = cv_model
        elif callbacks == None:
            best_model = cv_model
        save_model(
            best_model,
            "cross_validation/cv_best_model{}.h5".format(len(train_acc_scores) + 1),
        )
        pred = cv_model.predict(X_test, batch_size=batch_size, verbose=0)
        predictions = np.argmax(pred, axis=1)
        exec_time.append((end_time - start_time))
        train_acc_scores.append(
            cv_model.evaluate(X_train, Y_train, verbose=verbose)[1] * 100
        )
        acc_scores.append(
            accuracy_score(Y_origin[test].values, predictions, normalize=True) * 100
        )
        pres_scores.append(
            precision_score(Y_origin[test].values, predictions, average="macro") * 100
        )
        rec_scores.append(
            recall_score(Y_origin[test].values, predictions, average="macro") * 100
        )
        f1.append(f1_score(Y_origin[test].values, predictions, average="macro") * 100)
        cokap_scores.append(cohen_kappa_score(Y_origin[test].values, predictions) * 100)
        roc_auc_scores.append(roc_auc_score(Y_test.values, pred) * 100)
    scores = [
        "precision: %.4f%% (+/- %.4f%%)" % (np.mean(pres_scores), np.std(pres_scores)),
        "recall: %.4f%% (+/- %.4f%%)" % (np.mean(rec_scores), np.std(rec_scores)),
    ]
    cv_scores["metrics"] = [
        "preprocessing",
        "execution time",
        "training accuracy",
        "accuracy",
        "precision",
        "recall",
        "F1",
        "cohen_kappa",
        "roc_auc",
    ]
    _, model_name = model_fn.split("_", 1)
    cv_scores[model_name.replace("()", "")] = [
        normalizer_name,
        str(format(np.mean(exec_time), ".2f")) + " (s)",
        "%.4f%% (+/- %.4f%%)" % (np.mean(train_acc_scores), np.std(train_acc_scores)),
        "%.4f%% (+/- %.4f%%)" % (np.mean(acc_scores), np.std(acc_scores)),
        "%.4f%% (+/- %.4f%%)" % (np.mean(pres_scores), np.std(pres_scores)),
        "%.4f%% (+/- %.4f%%)" % (np.mean(rec_scores), np.std(rec_scores)),
        "%.4f%% (+/- %.4f%%)" % (np.mean(f1), np.std(f1)),
        "%.4f%% (+/- %.4f%%)" % (np.mean(cokap_scores), np.std(cokap_scores)),
        "%.4f%% (+/- %.4f%%)" % (np.mean(roc_auc_scores), np.std(roc_auc_scores)),
    ]
    cv_scores.to_csv(
        r"cross_validation/cross_validation_scores_{}.csv".format(normalizer_name),
        index=False,
    )
    return cv_scores


# ### learning rate visualization function
def plot_learning_rate(
    training_history, save=False, metric="accuracy", style="default"
):
    styles = [
        "seaborn-poster",
        "seaborn-bright",
        "Solarize_Light2",
        "seaborn-whitegrid",
        "_classic_test_patch",
        "seaborn-white",
        "fivethirtyeight",
        "seaborn-deep",
        "seaborn",
        "seaborn-dark-palette",
        "seaborn-paper",
        "seaborn-darkgrid",
        "seaborn-notebook",
        "grayscale",
        "seaborn-muted",
        "seaborn-dark",
        "seaborn-talk",
        "ggplot",
        "bmh",
        "dark_background",
        "fast",
        "seaborn-ticks",
        "seaborn-colorblind",
        "classic",
    ]
    if type(style) is int:
        try:
            style = styles[style]
        except Exception:
            print("back to default style, wrong style entry, choose style from 0 to 23")
            style = "default"
    if type(training_history) is pd.DataFrame:
        pass
    else:
        training_history = pd.DataFrame(training_history)
    pyplot.style.use(style)
    pyplot.figure(figsize=(20, 10))
    pyplot.title("{} ({} ephocs)".format(metric, len(training_history)))
    if metric == "all":
        for column in training_history.columns:
            if column != "Unnamed: 0":
                pyplot.plot(
                    [i for i in range(0, len(training_history))],
                    training_history[column],
                    label=str(column),
                )
                pyplot.legend()
                pyplot.grid(True)
                if save == True:
                    pyplot.savefig("plot.png")
    else:
        pyplot.plot(
            [i for i in range(0, len(training_history))],
            training_history["{}".format(metric)],
            label="train",
        )
        pyplot.plot(
            [i for i in range(0, len(training_history))],
            training_history["val_{}".format(metric)],
            label="test",
        )
        pyplot.legend()
        pyplot.grid(True)
        if save == True:
            pyplot.savefig("plot.png")



from maaml.Datasets.UAH_dataset import time_series
from maaml import preprocessing
from maaml import machine_learning
from maaml.functions.Datasets import time_series
from maaml.functions import preprocessing
from maaml.functions import machine_learning
from maaml.functions import deep_learning



import os
import pandas as pd


class PathBuilder:
    def __init__(
        self,
        dataset_dir=None,
        conditions_vector=[None, None, None, None],
        datatype=None,
        verbose=0,
    ):
        self.driver = conditions_vector[0]
        self.state = conditions_vector[1]
        self.roadtype = conditions_vector[2]
        self.selector = conditions_vector[3]
        self.verbose = verbose
        self.parent_dir = self.parent_path(dataset_dir, verbose=verbose)
        self.conditions = self.case_path_selection(
            self.driver, self.state, self.roadtype, self.selector, verbose=verbose
        )
        self.filetype = self.data_type_selection(datatype)
        if self.filetype != "" and self.conditions != "":
            self.path = os.path.join(self.parent_dir, self.conditions, self.filetype)
            if verbose == 1:
                print(f"\n your path is: {self.path}")
        else:
            self.path = ""
            if verbose == 1:
                print("\nError in the data type entry, empty path generated ")

    def parent_path(self, dataset_dir, verbose=1):
        self.dataset_dir = dataset_dir
        if dataset_dir == None:
            dataset_dir = input(
                "\nEnter the UAH dataset directory (example: '~/Downloads/UAH-DRIVESET-v1'), you can download it from here this link: \nhttp://www.robesafe.uah.es/personal/eduardo.romera/uah-driveset/#download \n\nyour input:"
            )
            if dataset_dir == "":
                print("\nno entry, loading default directory")
                PARENT_DIR = "UAH-DRIVESET-v1"
            else:
                PARENT_DIR = dataset_dir
        else:
            PARENT_DIR = str(dataset_dir)
        if verbose == 1:
            print(f"\nthe selected dataset directory is: {PARENT_DIR}")
        return PARENT_DIR

    def case_path_selection(self, driver, state, roadtype, selector=1, verbose=1):
        directory2 = None
        if driver == None:
            driver = input(
                "\nenter the driver from list: \n 1.'D1' 2.'D2' 3.'D3' 4.'D4' 5.'D5' 6.'D6' \n\nyour choice:"
            )
        else:
            driver = str(driver)
        if state == None:
            state = input(
                "\nenter state from list: \n 1.'normal' 2.'agressif' 3.'drowsy' \n\nyour choice:"
            )
        else:
            state = str(state)
        if roadtype == None:
            roadtype = input(
                "\nenter road type from list: \n 1.'secondary' 2.'motorway' \n\nyour choice:"
            )
        else:
            roadtype = str(roadtype)
        if driver == "1" or driver == "D1":
            if state == "1" or state == "normal":
                if roadtype == "1" or roadtype == "secondary":
                    directory = "D1/20151110175712-16km-D1-NORMAL1-SECONDARY"
                    directory2 = "D1/20151110180824-16km-D1-NORMAL2-SECONDARY"
                elif roadtype == "2" or roadtype == "motorway":
                    directory = "D1/20151111123124-25km-D1-NORMAL-MOTORWAY"
                else:
                    print(
                        "ERROR: bad or misspelled entry, try to write the entry correctly or remove the '', \n you can also use the entry numbers "
                    )
                    print("\n Reminder: we only have 2 road types")
            elif state == "2" or state == "agressif":
                if roadtype == "1" or roadtype == "secondary":
                    directory = "D1/20151111134545-16km-D1-AGGRESSIVE-SECONDARY"
                elif roadtype == "2" or roadtype == "motorway":
                    directory = "D1/20151111125233-24km-D1-AGGRESSIVE-MOTORWAY"
                else:
                    print(
                        "ERROR: bad or misspelled entry type, try to write the entry correctly or remove the '', \n you can also use the entry numbers "
                    )
                    print("\n Reminder: we only have 2 road types")
            elif state == "3" or state == "drowsy":
                if roadtype == "1" or roadtype == "secondary":
                    directory = "D1/20151111135612-13km-D1-DROWSY-SECONDARY"
                elif roadtype == "2" or roadtype == "motorway":
                    directory = "D1/20151111132348-25km-D1-DROWSY-MOTORWAY"
                else:
                    print(
                        "ERROR: bad or misspelled entry type, try to write the entry correctly or remove the '', \n you can also use the entry numbers "
                    )
                    print("\n Reminder: we only have 2 road types")
            else:
                print(
                    "ERROR: No such entry type or misspelled entry type, try to write the entry correctly or remove the '', \n you can also use the entry numbers "
                )
                print("\n Reminder: we only have 3 driving states")
        elif driver == "2" or driver == "D2":
            if state == "1" or state == "normal":
                if roadtype == "1" or roadtype == "secondary":
                    directory = "D2/20151120160904-16km-D2-NORMAL1-SECONDARY"
                    directory2 = "D2/20151120162105-17km-D2-NORMAL2-SECONDARY"
                elif roadtype == "2" or roadtype == "motorway":
                    directory = "D2/20151120131714-26km-D2-NORMAL-MOTORWAY"
                else:
                    print(
                        "ERROR: No such entry type or misspelled entry type, try to write the entry correctly or remove the '', \n you can also use the entry numbers "
                    )
                    print("\n Reminder: we only have 2 road types")
            elif state == "2" or state == "agressif":
                if roadtype == "1" or roadtype == "secondary":
                    directory = "D2/20151120163350-16km-D2-AGGRESSIVE-SECONDARY"
                elif roadtype == "2" or roadtype == "motorway":
                    directory = "D2/20151120133502-26km-D2-AGGRESSIVE-MOTORWAY"
                else:
                    print(
                        "ERROR: No such entry type or misspelled entry type, try to write the entry correctly or remove the '', \n you can also use the entry numbers "
                    )
                    print("\n Reminder: we only have 2 road types")
            elif state == "3" or state == "drowsy":
                if roadtype == "1" or roadtype == "secondary":
                    directory = "D2/20151120164606-16km-D2-DROWSY-SECONDARY"
                elif roadtype == "2" or roadtype == "motorway":
                    directory = "D2/20151120135152-25km-D2-DROWSY-MOTORWAY"
                else:
                    print(
                        "ERROR: No such entry type or misspelled entry type, try to write the entry correctly or remove the '', \n you can also use the entry numbers "
                    )
                    print("\n Reminder: we only have 2 road types")
            else:
                print(
                    "ERROR: No such entry type or misspelled entry type, try to write the entry correctly or remove the '', \n you can also use the entry numbers "
                )
                print("\n Reminder: we only have 3 driving states")
        elif driver == "3" or driver == "D3":
            if state == "1" or state == "normal":
                if roadtype == "1" or roadtype == "secondary":
                    directory = "D3/20151126124208-16km-D3-NORMAL1-SECONDARY"
                    directory2 = "D3/20151126125458-16km-D3-NORMAL2-SECONDARY"
                elif roadtype == "2" or roadtype == "motorway":
                    directory = "D3/20151126110502-26km-D3-NORMAL-MOTORWAY"
                else:
                    print(
                        "ERROR: No such entry type or misspelled entry type, try to write the entry correctly or remove the '', \n you can also use the entry numbers "
                    )
                    print("\n Reminder: we only have 2 road types")
            elif state == "2" or state == "agressif":
                if roadtype == "1" or roadtype == "secondary":
                    directory = "D3/20151126130707-16km-D3-AGGRESSIVE-SECONDARY"
                elif roadtype == "2" or roadtype == "motorway":
                    directory = "D3/20151126134736-26km-D3-AGGRESSIVE-MOTORWAY"
                else:
                    print(
                        "ERROR: No such entry type or misspelled entry type, try to write the entry correctly or remove the '', \n you can also use the entry numbers "
                    )
                    print("\n Reminder: we only have 2 road types")
            elif state == "3" or state == "drowsy":
                if roadtype == "1" or roadtype == "secondary":
                    directory = "D3/20151126132013-17km-D3-DROWSY-SECONDARY"
                elif roadtype == "2" or roadtype == "motorway":
                    directory = "D3/20151126113754-26km-D3-DROWSY-MOTORWAY"
                else:
                    print(
                        "ERROR: No such entry type or misspelled entry type, try to write the entry correctly or remove the '', \n you can also use the entry numbers "
                    )
                    print("\n Reminder: we only have 2 road types")
            else:
                print(
                    "ERROR: No such entry type or misspelled entry type, try to write the entry correctly or remove the '', \n you can also use the entry numbers "
                )
                print("\n Reminder: we only have 3 driving states")
        elif driver == "4" or driver == "D4":
            if state == "1" or state == "normal":
                if roadtype == "1" or roadtype == "secondary":
                    directory = "D4/20151203171800-16km-D4-NORMAL1-SECONDARY"
                    directory2 = "D4/20151203173103-17km-D4-NORMAL2-SECONDARY"
                elif roadtype == "2" or roadtype == "motorway":
                    directory = "D4/20151204152848-25km-D4-NORMAL-MOTORWAY"
                else:
                    print(
                        "ERROR: No such entry type or misspelled entry type, try to write the entry correctly or remove the '', \n you can also use the entry numbers "
                    )
                    print("\n Reminder: we only have 2 road types")
            elif state == "2" or state == "agressif":
                if roadtype == "1" or roadtype == "secondary":
                    directory = "D4/20151203174324-16km-D4-AGGRESSIVE-SECONDARY"
                elif roadtype == "2" or roadtype == "motorway":
                    directory = "D4/20151204154908-25km-D4-AGGRESSIVE-MOTORWAY"
                else:
                    print(
                        "ERROR: No such entry type or misspelled entry type, try to write the entry correctly or remove the '', \n you can also use the entry numbers "
                    )
                    print("\n Reminder: we only have 2 road types")
            elif state == "3" or state == "drowsy":
                if roadtype == "1" or roadtype == "secondary":
                    directory = "D4/20151203175637-17km-D4-DROWSY-SECONDARY"
                elif roadtype == "2" or roadtype == "motorway":
                    directory = "D4/20151204160823-25km-D4-DROWSY-MOTORWAY"
                else:
                    print(
                        "ERROR: No such entry type or misspelled entry type, try to write the entry correctly or remove the '', \n you can also use the entry numbers "
                    )
                    print("\n Reminder: we only have 2 road types")
            else:
                print(
                    "ERROR: No such entry type or misspelled entry type, try to write the entry correctly or remove the '', \n you can also use the entry numbers "
                )
                print("\n Reminder: we only have 3 driving states")
        elif driver == "5" or driver == "D5":
            if state == "1" or state == "normal":
                if roadtype == "1" or roadtype == "secondary":
                    directory = "D5/20151211162829-16km-D5-NORMAL1-SECONDARY"
                    directory2 = "D5/20151211164124-17km-D5-NORMAL2-SECONDARY"
                elif roadtype == "2" or roadtype == "motorway":
                    directory = "D5/20151209151242-25km-D5-NORMAL-MOTORWAY"
                else:
                    print(
                        "ERROR: No such entry type or misspelled entry type, try to write the entry correctly or remove the '', \n you can also use the entry numbers "
                    )
                    print("\n Reminder: we only have 2 road types")
            elif state == "2" or state == "agressif":
                if roadtype == "1" or roadtype == "secondary":
                    directory = "D5/20151211165606-12km-D5-AGGRESSIVE-SECONDARY"
                elif roadtype == "2" or roadtype == "motorway":
                    directory = "D5/20151209153137-25km-D5-AGGRESSIVE-MOTORWAY"
                else:
                    print(
                        "ERROR: No such entry type or misspelled entry type, try to write the entry correctly or remove the '', \n you can also use the entry numbers "
                    )
                    print("\n Reminder: we only have 2 road types")
            elif state == "3" or state == "drowsy":
                if roadtype == "1" or roadtype == "secondary":
                    directory = "D5/20151211170502-16km-D5-DROWSY-SECONDARY"
                elif roadtype == "2" or roadtype == "motorway":
                    directory = "D5/20151211160213-25km-D5-DROWSY-MOTORWAY"
                else:
                    print(
                        "ERROR: No such entry type or misspelled entry type, try to write the entry correctly or remove the '', \n you can also use the entry numbers "
                    )
                    print("\n Reminder: we only have 2 road types")
            else:
                print(
                    "ERROR: No such entry type or misspelled entry type, try to write the entry correctly or remove the '', \n you can also use the entry numbers "
                )
                print("\n Reminder: we only have 3 driving states")
        elif driver == "6" or driver == "D6":
            if state == "1" or state == "normal":
                if roadtype == "1" or roadtype == "secondary":
                    directory = "D6/20151221112434-17km-D6-NORMAL-SECONDARY"
                    directory2 = ""
                elif roadtype == "2" or roadtype == "motorway":
                    directory = "D6/20151217162714-26km-D6-NORMAL-MOTORWAY"
                else:
                    print(
                        "ERROR: No such entry type or misspelled entry type, try to write the entry correctly or remove the '' , \nyou can also use the entry numbers "
                    )
                    print("\nReminder: we only have 2 road types")
            elif state == "2" or state == "agressif":
                if roadtype == "1" or roadtype == "secondary":
                    directory = ""
                    if verbose == 1:
                        print(
                            "\nATTENTION! driver 6 does not have data for state :agressif and road type: secondary"
                        )
                elif roadtype == "2" or roadtype == "motorway":
                    directory = "D6/20151221120051-26km-D6-AGGRESSIVE-MOTORWAY"
                else:
                    print(
                        "ERROR: No such entry type or misspelled entry type, try to write the entry correctly or remove the '' , \nyou can also use the entry numbers "
                    )
                    print("\n Reminder: we only have 2 road types")
            elif state == "3" or state == "drowsy":
                if roadtype == "1" or roadtype == "secondary":
                    directory = "D6/20151221113846-16km-D6-DROWSY-SECONDARY"
                elif roadtype == "2" or roadtype == "motorway":
                    directory = "D6/20151217164730-25km-D6-DROWSY-MOTORWAY"
                else:
                    print(
                        "ERROR: No such entry type or misspelled entry type, try to write the entry correctly or remove the '' , \nyou can also use the entry numbers "
                    )
                    print("\nReminder: we only have 2 road types")
            else:
                print(
                    "ERROR: No such entry type or misspelled entry type, try to write the entry correctly or remove the '' , \nyou can also use the entry numbers "
                )
                print("\nReminder: we only have 3 driving states")
        else:
            directory = ""
            directory2 = ""
            print(
                "ERROR: No such entry type or misspelled entry type, try to write the entry correctly or remove the '' , \nyou can also use the entry numbers "
            )
            print("\nReminder: we only have 6 drivers")
        if selector == None:
            prompt = "We have two directories for this state and road type, choose which directory you want or you can choose to return both from the list: \n 1. 'directory1' 2.'directory2' 3.'both' \n\nyour choice:"
            selected = input(prompt)
        else:
            selected = str(selector)
        if (selected == "1" or selected == "directory 1") and directory != None:
            if verbose == 1:
                print(
                    f"you choose the default first directory: \nyour directory is: {directory}"
                )
            return directory
        elif (selected == "2" or selected == "directory2") and directory2 != None:
            if verbose == 1:
                print(
                    f"you choose the second directory: \nyour directory is: {directory2}"
                )
            return directory2
        elif (selected == "3" or selected == "both") and (
            directory != None and directory2 != None
        ):
            if verbose == 1:
                print(
                    f"you choose to select both directories: \nyour directories are: \n {directory}\n {directory2}\n"
                )
            return directory, directory2
        else:
            if verbose == 1:
                print(
                    "\nERROR: bad or misspelled entry, please change the entry while checking the spelling or removing the '' , \nyou can also use the entry numbers. \n"
                )
                print(
                    "Note: for Driver 6, the driver have a unique session in the case of normal state and secondary road \n"
                )
                print("Empty path generated")
            directory = ""
            return directory

    def data_type_selection(self, datatype):
        self.datatype = datatype
        SENSOR_FILE = [
            "RAW_GPS.txt",
            "RAW_ACCELEROMETERS.txt",
            "PROC_LANE_DETECTION.txt",
            "PROC_VEHICLE_DETECTION.txt",
            "PROC_OPENSTREETMAP_DATA.txt",
            "EVENTS_LIST_LANE_CHANGES.txt",
            "EVENTS_INERTIAL.txt",
            "SEMANTIC_FINAL.txt",
            "SEMANTIC_ONLINE.txt",
        ]
        if datatype == None:
            prompt = "\nplease choose which type of data from this list: \n 1.'GPS' 2.'Accelerometer' 3.'lane detection' 4.'vehicle detection' \n 5.'open street map' 6.'lane change events' 7.'inertial events' \n 8.'semantics final' 9.'semantics online' \n\nyour choice:"
            datatype = input(prompt)
        else:
            datatype = str(datatype)
        if datatype == "GPS" or datatype == "1":
            return SENSOR_FILE[0]
        elif datatype == "Accelerometer" or datatype == "2":
            return SENSOR_FILE[1]
        elif datatype == "lane detection" or datatype == "3":
            return SENSOR_FILE[2]
        elif datatype == "vehicle detection" or datatype == "4":
            return SENSOR_FILE[3]
        elif datatype == "open street map" or datatype == "5":
            return SENSOR_FILE[4]
        elif datatype == "lane change events" or datatype == "6":
            return SENSOR_FILE[5]
        elif datatype == "inertial events" or datatype == "7":
            return SENSOR_FILE[6]
        elif datatype == "semantics final" or datatype == "8":
            return SENSOR_FILE[7]
        elif datatype == "semantics online" or datatype == "9":
            return SENSOR_FILE[8]
        else:
            print(
                "ERROR: No such data type or misspelling, try to write it correctly or remove the '', \nyou can also use the data type numbers"
            )
            print("\nNo data type file name generated")
        return


class DataReader:
    def __init__(self, path, dataset_name=None, from_string=True, delimiter=" "):
        self.path = path
        if from_string is True:
            if dataset_name == "UAHdataset":
                if "SEMANTIC_FINAL" in path:
                    delimiter = None
                try:
                    self.data = pd.read_table(path, header=None, delimiter=delimiter)
                    self.data = self.uah_dataset_columns(path, self.data)
                except Exception:
                    print("\nERROR: please verify the entry path name for DataReader")
                    print("\nEmpty data table generated")
            else:
                self.data = pd.read_table(path, header=None, delimiter=delimiter)

    def uah_dataset_columns(self, path, data):
        self.path = path
        self.data
        SENSOR_FILE = [
            "RAW_GPS.txt",
            "RAW_ACCELEROMETERS.txt",
            "PROC_LANE_DETECTION.txt",
            "PROC_VEHICLE_DETECTION.txt",
            "PROC_OPENSTREETMAP_DATA.txt",
            "EVENTS_LIST_LANE_CHANGES.txt",
            "EVENTS_INERTIAL.txt",
            "SEMANTIC_FINAL.txt",
            "SEMANTIC_ONLINE.txt",
        ]
        if SENSOR_FILE[0] in path:
            data_column_names = [
                "Timestamp (seconds)",
                "Speed (km/h)",
                "Latitude coordinate (degrees)",
                "Longitude coordinate (degrees)",
                "Altitude (meters)",
                "Vertical accuracy (degrees)",
                "Horizontal accuracy (degrees)",
                "Course (degrees)",
                "Difcourse: course variation (degrees)",
            ]
            data = data.drop([9, 10, 11, 12], axis=1)
        elif SENSOR_FILE[1] in path:
            data_column_names = [
                "Timestamp (seconds)",
                "Boolean of system activated (1 if >50km/h)",
                "Acceleration in X (Gs)",
                "Acceleration in Y (Gs)",
                "Acceleration in Z (Gs)",
                "Acceleration in X filtered by KF (Gs)",
                "Acceleration in Y filtered by KF (Gs)",
                "Acceleration in Z filtered by KF (Gs)",
                "Roll (degrees)",
                "Pitch (degrees)",
                "Yaw (degrees)",
            ]
            try:
                data = data.drop(11, axis=1)
            except Exception:
                data = data
        elif SENSOR_FILE[2] in path:
            data_column_names = [
                "Timestamp (seconds)",
                "X: car position relative to lane center (meters)",
                "Phi: car angle relative to lane curvature (degrees)",
                "W: road width (meters)",
                "State of the lane det. algorithm [-1=calibrating,0=initializing, 1=undetected, 2=detected/running]",
            ]
        elif SENSOR_FILE[3] in path:
            data_column_names = [
                "Timestamp (seconds)",
                "Distance to ahead vehicle in current lane (meters) [value -1 means no car is detected in front]",
                "Time of impact to ahead vehicle (seconds) [distance related to own speed]",
                "Number of detected vehicles in this frame (traffic)",
                "GPS speed (km/h) [same as in RAW GPS]",
            ]
            data = data.drop(5, axis=1)
        elif SENSOR_FILE[4] in path:
            data_column_names = [
                "Timestamp (seconds)",
                "Maximum allowed speed of current road (km/h)",
                "Reliability of obtained maxspeed (0=unknown,1=reliable, 2=used previously obtained maxspeed,3=estimated by type of road)",
                "Type of road (motorway, trunk, secondary...)",
                "Number of lanes in current road",
                "Estimated current lane (1=right lane, 2=first left lane, 3=second left lane, etc) [experimental]",
                "GPS Latitude used to query OSM (degrees)",
                "GPS Longitude used to query OSM (degrees)",
                "OSM delay to answer query (seconds)",
                "GPS speed (km/h) [same as in RAW GPS]",
            ]
            data = data.drop(10, axis=1)
        elif SENSOR_FILE[5] in path:
            data_column_names = [
                "Timestamp (seconds)",
                "Type [+ indicates right and - left, 1 indicates normal lane change and 2 slow lane change]",
                "GPS Latitude of the event (degrees)",
                "GPS Longitude of the event (degrees)",
                "Duration of the lane change (seconds) [measured since the car position is near the lane marks]",
                "Time threshold to consider irregular change (secs.) [slow if change duration is over this threshold and fast if duration is lower than threshold/3]",
            ]
        elif SENSOR_FILE[6] in path:
            data_column_names = [
                "Timestamp (seconds)",
                "Type (1=braking, 2=turning, 3=acceleration)",
                "Level (1=low, 2=medium, 3=high)",
                "GPS Latitude of the event",
                "GPS Longitude of the event ",
                "Date of the event in YYYYMMDDhhmmss format",
            ]
            data = data.drop(6, axis=1)
        elif SENSOR_FILE[7] in path:
            data_column_names = [
                "Hour of route start",
                "Minute of route start",
                "Second of route start",
                "Average speed during trip (km/h)",
                "Maximum achieved speed during route (km/h)",
                "Lanex score (internal value, related to lane drifting)",
                "Driving time (in minutes)",
                "Hour of route end",
                "Minute of route end",
                "Second of route end",
                "Trip distance (km)",
                "ScoreLongDist	(internal value, Score accelerations)",
                "ScoreTranDist	(internal value, Score turnings)",
                "ScoreSpeedDist (internal value, Score brakings)",
                "ScoreGlobal (internal value, old score that is not used anymore)",
                "Alerts Long (internal value)",
                "Alerts Late (internal value)",
                "Alerts Lanex (internal value)",
                "Number of vehicle stops during route (experimental, related to fuel efficiency estimation)",
                "Speed variability (experimental, related to fuel efficiency estimation)",
                "Acceleration noise (experimental, related to fuel efficiency estimation)",
                "Kinetic energy (experimental, related to fuel efficiency estimation)",
                "Driving time (in seconds)",
                "Number of curves in the route",
                "Power exherted (experimental, related to fuel efficiency estimation)",
                "Acceleration events (internal value)",
                "Braking events (internal value)",
                "Turning events (internal value)",
                "Longitudinal-distraction Global Score (internal value, combines mean[31] and std[32])",
                "Transversal-distraction Global Score (internal value, combines mean[33] and std[34])",
                "Mean Long.-dist. score (internal value)",
                "STD Long.-dist. score (internal value)",
                "Average Trans.-dist. score (internal value)",
                "STD Trans.-dist. score (internal value)",
                "Lacc (number of low accelerations)",
                "Macc (number of medium accelerations)",
                "Hacc (number of high accelerations)",
                "Lbra (number of low brakings)",
                "Mbra (number of medium brakings)",
                "Hbra (number of high brakings)",
                "Ltur (number of low turnings)",
                "Mtur (number of medium turnings)",
                "Htur (number of high turnings)",
                "Score total (base 100, direct mean of the other 7 scores [45-51])",
                "Score accelerations (base 100)",
                "Score brakings (base 100)",
                "Score turnings (base 100)",
                "Score lane-weaving (base 100)",
                "Score lane-drifting (base 100)",
                "Score overspeeding (base 100)",
                "Score car-following (base 100)",
                "Ratio normal (base 1)",
                "Ratio drowsy (base 1)",
                "Ratio aggressive (base 1)",
            ]
            data = data.transpose()
        elif SENSOR_FILE[8] in path:
            data_column_names = [
                "TimeStamp since route start (seconds)",
                "GPS Latitude (degrees)",
                "GPS Longitude (degrees)",
                "Score total WINDOW (base 100, direct mean of the other 7 scores)",
                "Score accelerations WINDOW (base 100)",
                "Score brakings WINDOW (base 100)",
                "Score turnings WINDOW (base 100)",
                "Score weaving WINDOW (base 100)",
                "Score drifting WINDOW (base 100)",
                "Score overspeeding WINDOW (base 100)",
                "Score car-following WINDOW (base 100)",
                "Ratio normal WINDOW (base 1)",
                "Ratio drowsy WINDOW (base 1)",
                "Ratio aggressive WINDOW (base 1)",
                "Ratio distracted WINDOW (1=distraction detected in last 2 seconds, 0=otherwise)",
                "Score total (base 100, direct mean of the other 7 scores)",
                "Score accelerations (base 100)",
                "Score brakings (base 100)",
                "Score turnings (base 100)",
                "Score weaving (base 100)",
                "Score drifting (base 100)",
                "Score overspeeding (base 100)",
                "Score car-following (base 100)",
                "Ratio normal (base 1)",
                "Ratio drowsy (base 1)",
                "Ratio aggressive (base 1)",
                "Ratio distracted (1=distraction detected in last 2 seconds, 0=otherwise)",
            ]
            data = data.drop(27, axis=1)
        else:
            data_column_names = []
            print(
                "ERROR: No such file type or file name misspelled, try to write it correctly or remove the '' "
            )
            print("\nNo column names created\n")
        data.columns = data_column_names
        return data


class DataCleaner:
    def __init__(
        self,
        data,
        new_data=None,
        average_window=True,
        window_size=0,
        step=0,
        uah_dataset_vector: list = None,
        save_dataset=False,
        name_dataset="dataset",
        timestamp_column="Timestamp (seconds)",
        verbose=0,
    ):
        self.data_raw = data
        self.data_filtered = data.drop_duplicates(subset=[timestamp_column])
        self.average_window = average_window
        self.window_size = window_size
        self.step = step
        self.data_windowed = self.window_stepping(
            self.data_filtered,
            average_window=average_window,
            window_size=window_size,
            step=step,
            verbose=verbose,
        )
        if new_data is not None:
            self.data_merged = self.dataframes_merging(
                self.data_windowed,
                new_data,
                timestamp_column=timestamp_column,
                drop_duplicates=average_window,
                verbose=verbose,
            )
        else:
            self.data_merged = self.data_windowed
        self.data_interpolated = self.data_interpolating(
            self.data_merged, timestamp_columns=timestamp_column, verbose=verbose
        )
        self.dataset = self.removing_incomplete_raws(
            self.data_interpolated, verbose=verbose
        )
        if uah_dataset_vector is not None:
            self.dataset = self.column_adding(
                self.dataset,
                uah_dataset_vector=uah_dataset_vector,
                verbose=verbose,
            )
        if save_dataset == True:
            NEW_PATH = "dataset"
            if not os.path.exists(NEW_PATH):
                os.makedirs(NEW_PATH)
            self.dataset.to_csv(f"{NEW_PATH}/{name_dataset}.csv", index=False)

    @staticmethod
    def window_stepping(data=[], window_size=0, step=0, average_window=True, verbose=1):
        segment = []
        final_data = pd.DataFrame()
        if len(data) != 0:
            if window_size == 0:
                final_data = data
                if verbose == 1:
                    print("\nATTENTION: Entry data returned without window stepping")
                return final_data
            else:
                if average_window is True:
                    if verbose == 1:
                        print("\nAverage window applied")
                    for i in range(0, len(data) - 1, step):
                        segment = data[i : i + window_size]
                        row = segment.mean()
                        final_data = final_data.append(row, ignore_index=True)
                else:
                    for i in range(0, len(data) - 1, step):
                        window = data[i : i + window_size]
                        final_data = final_data.append(window, ignore_index=True)
                    if verbose == 1:
                        print(
                            f"\nwindow stepping applied with window size: {window_size} and step : {step}"
                        )
        else:
            final_data = []
            print("ERROR: Empty data entry")
        return final_data

    @staticmethod
    def dataframes_merging(
        data=[],
        new_data=[],
        timestamp_column="Timestamp (seconds)",
        drop_duplicates=True,
        verbose=1,
    ):
        try:
            while data.dtypes[timestamp_column] != "int64":
                if verbose == 1:
                    print(
                        "\nWarning: data Timestamp type is: ",
                        data.dtypes[timestamp_column],
                        "\n",
                    )
                data = data.astype({timestamp_column: "int"})
                if verbose == 1:
                    print(
                        "data timestamp type changed to : ",
                        data.dtypes[timestamp_column],
                        "\n",
                    )
            while new_data.dtypes[timestamp_column] != "int64":
                if verbose == 1:
                    print(
                        "Warning: new_data Timestamp type is: ",
                        data.dtypes[timestamp_column],
                        "\n",
                    )
                new_data = new_data.astype({timestamp_column: "int"})
                if verbose == 1:
                    print(
                        "new_data timestamp type changed to : ",
                        new_data.dtypes[timestamp_column],
                        "\n",
                    )
            if drop_duplicates is True:
                data = data.drop_duplicates([timestamp_column])
                new_data = new_data.drop_duplicates([timestamp_column])
            data_merged = data.set_index(timestamp_column).join(
                new_data.set_index(timestamp_column)
            )
            data_merged = data_merged.reset_index()
            if verbose == 1:
                print(f"Shape of the megred data: {data_merged.shape}\n")
                print("\033[1m", "******* DATA SUCCESSFULLY MERGED *******", "\033[0m")
        except Exception:
            print(
                "ERROR: empty data entries or one data entry or both do not have Timestamp column, \nplease renter your two dataframes and check their columns before entry "
            )
            print("\nEmpty data returned")
            data_merged = []
        return data_merged

    @staticmethod
    def data_interpolating(
        data=[], timestamp_columns=["Timestamp (seconds)"], verbose=1
    ):
        try:
            if verbose == 1:
                print(
                    f"\n    State before interpolation    \nCOLUMNS                   NUMBER OF RAWS WITH MISSING DATA\n{data.isnull().sum()}\n"
                )
            if data.isnull().values.any() == True:
                if verbose == 1:
                    print("\n       Executing interpolation     \n")
                missing_values = data.drop(timestamp_columns, axis=1)
                missing_values = missing_values.interpolate(method="cubic", limit=3)
                data[missing_values.columns] = missing_values
                data_interpolated = data
                if verbose == 1:
                    print(
                        f"\n    State after interpolation    \nCOLUMNS                   NUMBER OF RAWS WITH MISSING DATA\n{data_interpolated.isnull().sum()}\n"
                    )
            else:
                data_interpolated = data
                if verbose == 1:
                    print("\n   Interpolation not needed    \n")
        except Exception:
            data_interpolated = []
            print(
                f"{data_interpolated}\nERROR: empty data entry or non dataframe type\nEmpty data returned"
            )
        return data_interpolated

    @staticmethod
    def removing_incomplete_raws(data=[], verbose=1):
        try:
            if verbose == 1:
                print(
                    f"\n    Data count before removing any rows :     \n{data.count()}"
                )
                print(
                    "\nis there any missing data values? :",
                    "\033[1m",
                    data.isnull().values.any(),
                    "\033[0m",
                )
            data = data.dropna()
            data = data.reset_index(drop=True)
            if verbose == 1:
                print(f"\n  Final Data count :     \n{data.count()}")
                print(
                    "\nis there any missing data values? :",
                    "\033[1m",
                    data.isnull().values.any(),
                    "\033[0m",
                )
        except Exception:
            print(
                "ERROR: empty data entry or non dataframe type, please enter your data dataframe\nEmpty data returned"
            )
            data = []
        return data

    @staticmethod
    def column_adding(
        data,
        column_name: str = None,
        value: str = None,
        uah_dataset_vector: list = None,
        verbose=0,
    ):
        if uah_dataset_vector is None:
            if column_name is not None and value is not None:
                data[column_name] = value
            else:
                data = data
                if verbose == 1:
                    print("\n    No label columns added    \n")
        else:
            if uah_dataset_vector[0] == 1:
                data["driver"] = "1"
            elif uah_dataset_vector[0] == 2:
                data["driver"] = "2"
            elif uah_dataset_vector[0] == 3:
                data["driver"] = "3"
            elif uah_dataset_vector[0] == 4:
                data["driver"] = "4"
            elif uah_dataset_vector[0] == 5:
                data["driver"] = "5"
            elif uah_dataset_vector[0] == 6:
                data["driver"] = "6"
            if uah_dataset_vector[2] == 1:
                data["road"] = "secondary"
            elif uah_dataset_vector[2] == 2:
                data["road"] = "motorway"
            if uah_dataset_vector[1] == 1:
                data["target"] = "normal"
            elif uah_dataset_vector[1] == 2:
                data["target"] = "agressif"
            elif uah_dataset_vector[1] == 3:
                data["target"] = "drowsy"
            if verbose == 1:
                print("\n   labels columns added successfully   \n")
        return data


class UAHDatasetBuilder:
    def __init__(
        self,
        path,
        datatype1,
        datatype2,
        window_size_dt1=0,
        step_dt1=0,
        window_size_dt2=0,
        step_dt2=0,
        save_dataset=False,
        name_dataset="UAHDataset",
        verbose=0,
        verbose1=0,
        verbose2=0,
    ):
        self.path_list = []
        self.path_list2 = []
        self.data = pd.DataFrame()
        for i in (1, 2, 3, 4, 5, 6):
            for j in (1, 2, 3):
                for k in (1, 2):
                    for l in (1, 2):
                        file1 = PathBuilder(
                            path, conditions_vector=[i, j, k, l], datatype=datatype1
                        )
                        file2 = PathBuilder(
                            path, conditions_vector=[i, j, k, l], datatype=datatype2
                        )
                        if file1.path != "" and file2.path != "":
                            raw_data1 = DataReader(
                                file1.path, dataset_name="UAHdataset"
                            )
                            raw_data2 = DataReader(
                                file2.path, dataset_name="UAHdataset"
                            )
                            self.data_chunk1 = DataCleaner(
                                raw_data1.data,
                                window_size=window_size_dt1,
                                step=step_dt1,
                                verbose=verbose1,
                            ).dataset
                            self.data_chunk2 = DataCleaner(
                                raw_data2.data,
                                window_size=window_size_dt2,
                                step=step_dt2,
                                verbose=verbose2,
                            ).dataset
                            self.data_chunk_merged = DataCleaner(
                                data=self.data_chunk1,
                                new_data=self.data_chunk2,
                                uah_dataset_vector=[i, j, k, l],
                                verbose=verbose,
                            ).dataset
                            self.data = self.data.append(self.data_chunk_merged)
                            self.data = self.data.reset_index()
                            self.data = self.data.drop("index", axis=1)
                            self.path_list.append(file1.path)
                            self.path_list2.append(file2.path)
        if save_dataset == True:
            NEW_PATH = "dataset"
            if not os.path.exists(NEW_PATH):
                os.makedirs(NEW_PATH)
            self.data.to_csv(f"{NEW_PATH}/{name_dataset}.csv", index=False)
            if verbose == 1:
                print(
                    f"\n\033[1m++++++++++++ UAHDATASET IS SAVED IN FILE : {os.getcwd()}/{NEW_PATH}/{name_dataset}.csv ++++++++++++\033[0m\n"
                )


if __name__ == "__main__":
    DATA_DIR_PATH = "/run/media/najem/34b207a8-0f0c-4398-bba2-f31339727706/home/stock/The_stock/dev & datasets/PhD/datasets/UAH-DRIVESET-v1/"
    dataset = UAHDatasetBuilder(
        DATA_DIR_PATH,
        1,
        2,
        window_size_dt2=10,
        step_dt2=10,
        verbose=1,
        save_dataset=True,
    )
    print(f"The dataset shape: {dataset.data.shape}")



from maaml.functions.preprocessing import (
    dataset_loader,
    sklearn_data_normalizer,
    encoding_label,
)
import numpy as np
import pandas as pd
import time
from sklearn.model_selection import train_test_split, ShuffleSplit
from sklearn.preprocessing import label_binarize
from sklearn.metrics import (
    accuracy_score,
    precision_score,
    recall_score,
    f1_score,
    cohen_kappa_score,
    roc_auc_score,
    classification_report,
)
from sklearn.tree import DecisionTreeClassifier
from sklearn.neighbors import KNeighborsClassifier
from sklearn.ensemble import RandomForestClassifier
from sklearn.tree import ExtraTreeClassifier
from sklearn.ensemble import ExtraTreesClassifier
from sklearn.naive_bayes import GaussianNB
from sklearn import svm
from sklearn.linear_model import LogisticRegression
from sklearn.neural_network import MLPClassifier

# function blocks
# # model selection function
def createMLmodel(model_name="", paramater=100, verbose=0):
    model_name = str(model_name)
    if model_name == "1" or model_name == "DecisionTree":
        model = DecisionTreeClassifier()
        model_name = "DecisionTree"
    elif model_name == "2" or model_name == "RandomForest":
        paramater = int(paramater)
        model = RandomForestClassifier(n_estimators=paramater)
        model_name = "RandomForest"
    elif model_name == "3" or model_name == "ExtraTree":
        model = ExtraTreeClassifier()
        model_name = "ExtraTree"
    elif model_name == "4" or model_name == "ExtraTrees":
        paramater = int(paramater)
        model = ExtraTreesClassifier(n_estimators=paramater)
        model_name = "ExtraTrees"
    elif model_name == "5" or model_name == "KNeighbors":
        paramater = int(paramater)
        model = KNeighborsClassifier(n_neighbors=paramater)
        model_name = "KNeighbors"
    elif model_name == "6" or model_name == "GaussianNB":
        model = GaussianNB()
        model_name = "GaussianNB"
    elif model_name == "7" or model_name == "svm":
        paramater = str(paramater)
        model = svm.SVC(gamma=paramater)
        model_name = "svm"
    elif model_name == "8" or model_name == "LogisticRegression":
        paramater = str(paramater)
        model = LogisticRegression(solver=paramater, multi_class="auto", max_iter=1000)
        model_name = "LogisticRegression"
    elif model_name == "9" or model_name == "MLPClassifier":
        paramater = int(paramater)
        model = MLPClassifier(max_iter=paramater)
        model_name = "MLPClassifier"
    else:
        print(
            'ERROR:wrong entry, "createMLmodel" have 9 diffrent classifiers, you could choose by number or by name'
        )
        model_name = "Empty model"
        model = None
    if verbose == 1:
        print(model_name, "selected")
    return model, model_name


# # cross validation function
def cross_validate_model(
    dataset,
    target_name="target",
    model_name="",
    model_param=100,
    nb_splits=5,
    test_size=0.3,
    verbose=0,
):
    X = dataset.drop(target_name, axis=1)
    Y = dataset[target_name]
    cv = ShuffleSplit(n_splits=nb_splits, test_size=test_size, random_state=10)
    acc_scores, pres_scores, rec_scores, f1, cokap_scores, roc_auc_scores = (
        [],
        [],
        [],
        [],
        [],
        [],
    )
    cv_scores = []
    for train, test in cv.split(X, Y):
        classes = Y.unique()
        y_testb = label_binarize(Y[test], classes=classes)
        model, model_name = createMLmodel(model_name=model_name, paramater=model_param)
        model = model.fit(X.loc[train], Y[train]).predict(X.loc[test])
        acc_scores.append(accuracy_score(Y[test], model, normalize=True) * 100)
        pres_scores.append(precision_score(Y[test], model, average="macro") * 100)
        rec_scores.append(recall_score(Y[test], model, average="macro") * 100)
        f1.append(f1_score(Y[test], model, average="macro") * 100)
        cokap_scores.append(cohen_kappa_score(Y[test], model) * 100)
        roc_auc_scores.append(roc_auc_score(y_testb, model.reshape(-1, 1)) * 100)
    cv_scores = [
        model_name,
        "accuracy: %.2f%% (+/- %.2f%%)" % (np.mean(acc_scores), np.std(acc_scores)),
        "precision: %.2f%% (+/- %.2f%%)" % (np.mean(pres_scores), np.std(pres_scores)),
        "recall: %.2f%% (+/- %.2f%%)" % (np.mean(rec_scores), np.std(rec_scores)),
        "F1: %.2f%% (+/- %.2f%%)" % (np.mean(f1), np.std(f1)),
        "cohen_kappa: %.2f%% (+/- %.2f%%)"
        % (np.mean(cokap_scores), np.std(cokap_scores)),
        "roc_auc: %.2f%% (+/- %.2f%%)"
        % (np.mean(roc_auc_scores), np.std(roc_auc_scores)),
    ]
    if verbose == 1:
        print("\033[1m" + "\n" + model_name + ": " + "\033[0m")
        print(
            "accuracy cross validation :",
            "%.4f%% (+/- %.4f%%)" % (np.mean(acc_scores), np.std(acc_scores)),
        )
        print(
            "precision cross validation :",
            "%.4f%% (+/- %.4f%%)" % (np.mean(pres_scores), np.std(pres_scores)),
        )
        print(
            "recall cross validation :",
            "%.4f%% (+/- %.4f%%)" % (np.mean(rec_scores), np.std(rec_scores)),
        )
        print(
            "f1 cross validation :", "%.4f%% (+/- %.4f%%)" % (np.mean(f1), np.std(f1))
        )
        print(
            "cohen_kappa cross validation :",
            "%.4f%% (+/- %.4f%%)" % (np.mean(cokap_scores), np.std(cokap_scores)),
        )
        print(
            "roc_auc_score cross validation :",
            "%.4f%% (+/- %.4f%%)" % (np.mean(roc_auc_scores), np.std(roc_auc_scores)),
        )
        print("\n")
    if verbose == 2:
        print("\n                          Accuracy Detailed splits :\n\n", acc_scores)
        print(
            "\n                          Precision Detailed splits :\n\n", pres_scores
        )
        print("\n                           Detailed splits accuracy:\n\n", rec_scores)
        print("\n                          F1 Detailed splits :\n\n", f1)
        print(
            "\n                          Cohen_kappa Detailed splits :\n\n",
            cokap_scores,
        )
        print(
            "\n                          Roc_Auc_score Detailed splits :\n\n",
            roc_auc_scores,
        )
    return cv_scores


# # feature importance function
def extract_features_importance(
    df, classifier="default", target_name="target", split_size=0.3, verbose=0
):
    x = df.drop(target_name, axis=1)
    y = df[target_name]
    X_train, X_test, y_train, y_test = train_test_split(
        x, y, test_size=split_size, random_state=10
    )
    if classifier == "default":
        model = RandomForestClassifier(n_estimators=100)
        print("Default Classifier is RandomForestClassifier ")
    else:
        model = classifier
    model = model.fit(X_train, y_train)
    pred = model.predict(X_test)

    if verbose == 1:
        print(
            str(classifier),
            "Classifier accuracy : ",
            accuracy_score(y_test, pred, normalize=True),
            "\n",
        )
        print(classification_report(y_test, pred))
    importances = model.feature_importances_
    ranks = pd.DataFrame(importances * 100, index=x.columns, columns=["importance %"])
    ranks = ranks.sort_values("importance %")[::-1]
    return ranks


# # full evaluation pipeline function
def evaluation_pipline(
    path,
    specific,
    droped_colummns=[],
    prediction_column="target",
    model_parameter=100,
    feature_importance=True,
    verbose=0,
):
    if specific == "all":
        data = dataset_loader(path=path, full=True)
    elif specific == "" or specific == "secondary road":
        data = dataset_loader(path=path, full=False, specific=specific, verbose=verbose)
    elif specific == "0" or specific == "motorway road":
        data = dataset_loader(path=path, full=False, specific=specific, verbose=verbose)
    elif int(specific) < 7:
        data = dataset_loader(path=path, full=False, specific=specific, verbose=verbose)
    for column_name in droped_colummns:
        data = data.drop(column_name, axis=1)
        if verbose == 1:
            print("droped colummn :", column_name)
    normalizer_result, importance, results = (
        pd.DataFrame(),
        pd.DataFrame(),
        pd.DataFrame(),
    )
    for i in range(0, 9):
        data_numeric = encoding_label(
            encoding_label(data, "target", verbose=verbose), "road", verbose=verbose
        ).drop("Timestamp (seconds)", axis=1)
        df, normalizer_name = sklearn_data_normalizer(
            data_numeric, [prediction_column], i, verbose=verbose
        )
        if verbose == 1:
            print("number of features used: ", df.shape[1] - 1)
        for m in range(1, 10):
            model_param = model_parameter
            if m == 7:
                model_param = "scale"
            elif m == 8:
                model_param = "lbfgs"
            elif m == 9:
                model_param = 2000
            start_time = time.time()
            cross_validation = cross_validate_model(
                df,
                model_name=m,
                model_param=model_param,
                target_name=prediction_column,
                nb_splits=5,
                verbose=verbose,
            )
            end_time = time.time()
            time_recorded = str(format((end_time - start_time) / 5, ".2f")) + " (s)"
            if verbose == 1:
                print(
                    createMLmodel(model_name=m, paramater=model_param)[1],
                    "excution time (single fold): ",
                    time_recorded,
                )
            index, model_results = ["preprocessing", "execution time"], [
                normalizer_name,
                time_recorded,
            ]
            for j in range(1, len(cross_validation)):
                ind, res = cross_validation[j].split(":")
                model_results.append(res)
                index.append(ind)
            normalizer_result["metrics"] = index
            normalizer_result[cross_validation[0]] = model_results
            if m < 5 and feature_importance == True:
                classifier = createMLmodel(model_name=m, paramater=model_param)[0]
                ranks = extract_features_importance(
                    df, classifier, target_name=prediction_column, verbose=0
                )
            if feature_importance == True:
                importance = importance.append(ranks)
        results = results.append(normalizer_result)
    if feature_importance == True:
        return results, importance
    return results
